from collections.abc import Awaitable, Mapping, Sequence
from contextlib import AbstractAsyncContextManager
from typing import (
    Any,
    Callable,
    Optional,
    Union,
)

from starlette.applications import AppType, Starlette
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import BaseRoute
from starlette.websockets import WebSocket

from .globals import GlobalsMiddleware
from .message_flash import MessageFlashMiddleware
from .middleware.user_sessions import UserSessionMiddleware
from .routing import AppRouter

RouteFunctionType = Callable[[Request], Union[Awaitable[Response], Response]]


class Mojito(Starlette):
    def __init__(
        self: AppType,
        debug: bool = False,
        routes: Optional[Sequence[BaseRoute]] = None,
        middleware: Optional[Sequence[Middleware]] = None,
        exception_handlers: Optional[
            Mapping[
                Any,
                Union[
                    Callable[
                        [Request, Exception], Union[Response, Awaitable[Response]]
                    ],
                    Callable[[WebSocket, Exception], Awaitable[None]],
                ],
            ]
        ] = None,
        on_startup: Optional[Sequence[Callable[[], Any]]] = None,
        on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
        lifespan: Optional[
            Union[
                Callable[[AppType], AbstractAsyncContextManager[None]],
                Callable[[AppType], AbstractAsyncContextManager[Mapping[str, Any]]],
            ]
        ] = None,
    ) -> None:
        super().__init__(
            debug,
            routes,
            middleware,
            exception_handlers,
            on_startup,
            on_shutdown,
            lifespan,
        )
        self.router = AppRouter(lifespan=lifespan)
        self.add_middleware(GlobalsMiddleware)
        self.add_middleware(UserSessionMiddleware)
        self.add_middleware(MessageFlashMiddleware)

    def include_router(self, router: AppRouter) -> None:
        """Mounts all the routers routes under the application with the prefix.

        Args:
            router (AppRouter): Instance of the AppRouter
        """
        self.router.include_router(router)  # type:ignore [attr-defined]

    def route(
        self,
        path: str,
        methods: Optional[list[str]] = ["GET"],
        name: Optional[str] = None,
        include_in_schema: bool = True,
    ) -> Callable[..., RouteFunctionType]:
        return self.router.route(
            path=path, methods=methods, name=name, include_in_schema=include_in_schema
        )



# Credit to https://gist.github.com/ddanier/ead419826ac6c3d75c96f9d89bea9bd0
from collections.abc import Awaitable, Callable
from contextvars import ContextVar, copy_context
from typing import Any

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp


class GlobalContextVar:
    __slots__ = ("_vars", "_defaults")

    _vars: dict[str, ContextVar[Any]]
    _defaults: dict[str, Any]

    def __init__(self) -> None:
        object.__setattr__(self, "_vars", {})
        object.__setattr__(self, "_defaults", {})

    def set_default(self, name: str, default: Any) -> None:
        """Set a default value for a variable."""

        # Ignore if default is already set and is the same value
        if name in self._defaults and default is self._defaults[name]:
            return

        # Ensure we don't have a value set already - the default will have
        # no effect then
        if name in self._vars:
            raise RuntimeError(
                f"Cannot set default as variable {name} was already set",
            )

        # Set the default already!
        self._defaults[name] = default

    def _get_default_value(self, name: str) -> Any:
        """Get the default value for a variable."""

        default = self._defaults.get(name, None)

        return default() if callable(default) else default

    def _ensure_var(self, name: str) -> None:
        """Ensure a ContextVar exists for a variable."""

        if name not in self._vars:
            default = self._get_default_value(name)
            self._vars[name] = ContextVar(f"globals:{name}", default=default)

    def __getattr__(self, name: str) -> Any:
        """Get the value of a variable."""

        self._ensure_var(name)
        return self._vars[name].get()

    def __setattr__(self, name: str, value: Any) -> None:
        """Set the value of a variable."""

        self._ensure_var(name)
        self._vars[name].set(value)


async def globals_middleware_dispatch(
    request: Request,
    call_next: Callable[..., Awaitable[Response]],
) -> Response:
    """Dispatch the request in a new context to allow globals to be used."""

    ctx = copy_context()

    def _call_next() -> Awaitable[Response]:
        return call_next(request)

    return await ctx.run(_call_next)


g = GlobalContextVar()
"""
Global context variable similar to Flasks g.

Usage:
    g.foo = "Foo"

    print(g.foo) -> "Foo"
"""


class GlobalsMiddleware(BaseHTTPMiddleware):  # noqa
    """Middleware to setup the globals context using globals_middleware_dispatch().

    Should be the first middleware processed in the application. Sets g.request to the
    current request."""

    def __init__(self, app: ASGIApp) -> None:
        super().__init__(app, globals_middleware_dispatch)



import asyncio

import httpx

from mojito import Mojito, Request, g

app = Mojito()


@app.route("/{g_message}")
async def get_and_set_g(request: Request, g_message: int):
    g.test_message = g_message
    await asyncio.sleep(1)
    assert g.test_message == g_message


async def get_g(i: int):
    transport = httpx.ASGITransport(app=app)
    async with httpx.AsyncClient(transport=transport, base_url="http://test") as client:
        await client.get(f"/{i}")


async def main():
    await asyncio.gather(*[get_g(i) for i in range(1000, 2000)])


def test_g_isolation():
    asyncio.run(main())
    assert True


if __name__ == "__main__":
    asyncio.run(main())



import inspect
from collections.abc import Awaitable, Mapping, Sequence
from typing import Any, Callable, Optional, Union

from starlette.applications import P
from starlette.background import BackgroundTask
from starlette.datastructures import URL
from starlette.middleware import (
    Middleware,
    _MiddlewareClass,  # type: ignore [unused-ignore]
)
from starlette.requests import Request
from starlette.responses import HTMLResponse, RedirectResponse, Response
from starlette.routing import PARAM_REGEX, BaseRoute, Route, Router
from starlette.types import AppType, Lifespan

from . import helpers
from .config import Config
from .globals import g

RouteFunctionType = Callable[[Request], Union[Awaitable[Response], Response]]


class AppRouter(Router):
    def __init__(
        self,
        prefix: Optional[str] = None,
        name: Optional[str] = None,
        middleware: Optional[Sequence[Middleware]] = None,
        routes: Optional[list[BaseRoute]] = None,
        lifespan: Optional[Lifespan[AppType]] = None,
    ) -> None:
        super().__init__(routes=routes, lifespan=lifespan)
        self.middleware = [] if middleware is None else list(middleware)
        self.prefix = prefix or ""
        self.routes: list[BaseRoute] = []
        self.name = name if name else ""

    def include_router(self, router: "AppRouter") -> None:
        for route in router.routes:
            self.routes.append(route)

    def add_middleware(
        self,
        middleware_class: type[_MiddlewareClass[P]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """Adds includes middleware within the router.

        Args:
            middleware_class (type[_MiddlewareClass[P]]): ASGI compatible middleware
        """
        self.middleware.insert(0, Middleware(middleware_class, *args, **kwargs))

    def add_route(
        self,
        path: str,
        endpoint: RouteFunctionType,
        methods: Optional[list[str]] = None,
        name: Optional[str] = None,
        include_in_schema: bool = True,
    ) -> None:
        path = self.prefix + path
        self.routes.append(
            Route(
                path=path,
                endpoint=endpoint,
                methods=methods,
                name=name,
                include_in_schema=include_in_schema,
                middleware=self.middleware,  # Added to the route if not mounting
            )
        )

    def _process_endpoint_args(
        self, request: Request, path: str, endpoint_function: Callable[..., Any]
    ) -> dict[str, Any]:
        """Inspect the endpoint function arguments and inject their values as **kwargs when
        the function is called from the endpoint.

        Args:
            request (Request): starlette.Request
            path (str): route path
            endpoint_function (Callable[..., Any]): The endpoint function

        Returns:
            dict[str, Any]: kwargs to call the endpoint function with
        """
        path_params_tuple = PARAM_REGEX.findall(
            path
        )  # Pull the path params from the path with converter split if it exists
        arg_specs = inspect.getfullargspec(endpoint_function)
        path_params = [p[0] for p in path_params_tuple]  # Extract just param name
        kwargs: dict[str, Any] = {}
        for arg_name, arg_type in arg_specs.annotations.items():
            arg_value: Optional[Any] = None
            if arg_name in path_params:
                # Arguments in path params
                arg_value = request.path_params.get(arg_name)
            elif arg_name == "return":
                # Handle 'return' type if one is included.
                # https://docs.python.org/3/library/inspect.html#inspect.getfullargspec
                # Skip processing 'return' type annotation
                continue
            elif arg_type == Request:
                # Add request to this argument
                arg_value = request
            else:
                # Remaining arguments are treated as query params
                arg_value = request.query_params.get(arg_name)
            kwargs[arg_name] = arg_value
        return kwargs

    def route(
        self,
        path: str,
        methods: Optional[list[str]] = ["GET"],
        name: Optional[str] = None,
        include_in_schema: bool = True,
    ) -> Callable[[Callable[..., Any]], RouteFunctionType]:
        def decorator(
            func: Callable[..., Union[Awaitable[Any], Any]],
        ) -> RouteFunctionType:
            async def endpoint_function(request: Request) -> Response:
                """Creates a function that inputs the correct arguments to the func at runtime."""
                kwargs = self._process_endpoint_args(request, path, func)
                g.request = request

                # Ensures the function has a Response return type.
                original_response: Any = func(**kwargs)
                if isinstance(original_response, Awaitable):
                    original_response = await original_response
                if not isinstance(original_response, Response):
                    # Wrap response in default HTMLResponse
                    original_response = HTMLResponse(str(original_response))
                response: Response = original_response
                # PROCESS MESSAGE FLASHING FOR NEXT REQUEST
                if g.next_flash_messages:
                    response.set_cookie(
                        Config.MESSAGE_FLASH_COOKIE,
                        helpers.encode_message_cookie(g.next_flash_message).decode(
                            "utf-8"
                        ),
                    )
                return response

            self.add_route(
                path,
                endpoint_function,
                methods=methods,
                name=name if name else func.__name__,
                include_in_schema=include_in_schema,
            )
            return func

        return decorator


def redirect_to(
    url: Union[str, URL],
    status_code: int = 302,
    headers: Optional[Mapping[str, str]] = None,
    background: Optional[BackgroundTask] = None,
) -> RedirectResponse:
    return RedirectResponse(url, status_code, headers, background)



from starlette.responses import FileResponse as FileResponse  # noqa
from starlette.responses import HTMLResponse as HTMLResponse  # noqa
from starlette.responses import JSONResponse as JSONResponse  # noqa
from starlette.responses import PlainTextResponse as PlainTextResponse  # noqa
from starlette.responses import RedirectResponse as RedirectResponse  # noqa
from starlette.responses import Response as Response  # noqa
from starlette.responses import StreamingResponse as StreamingResponse  # noqa



from mojito import AppRouter, Mojito, Request, auth, config
from mojito.testclient import TestClient

from .db import get_db

config.Config.SUPERUSER_PERMISSION_NAME = "superuser"
app = Mojito()
protected_router = AppRouter()
protected_router.add_middleware(auth.AuthMiddleware)

client = TestClient(app)


class PasswordAuth(auth.BaseAuth):
    "Authenticate with username and password"

    async def authenticate(self, request: Request, **kwargs: dict[str, str]):
        email: str = kwargs.get("username")  # type:ignore
        password: str = kwargs.get("password")  # type:ignore
        assert email
        assert password
        async with get_db() as db:
            user = await (
                await db.execute(f"SELECT * FROM users where email = '{email}'")
            ).fetchone()
        if not user:
            raise ValueError("No user found in database")
        if not auth.hash_password(password) == user["password"]:
            return None
        user_dict = dict(user)
        del user_dict["password"]
        auth_data = auth.AuthSessionData(
            is_authenticated=True,
            auth_handler="PasswordAuth",
            user_id=user["id"],
            data=dict(user),
            permissions=["admin"],
        )
        return auth_data

    async def get_user(self, user_id: int) -> auth.AuthSessionData:
        async with get_db() as db:
            user = await (
                await db.execute(
                    f"SELECT id, name, email, is_active FROM users where id = {user_id}"
                )
            ).fetchone()
        if not user:
            raise ValueError("No user found in database")
        return auth.AuthSessionData(
            is_authenticated=True,
            auth_handler="PasswordAuth",
            user_id=user["id"],
            data=dict(user),
            permissions=["admin"],
        )


class SuperuserPasswordAuth(auth.BaseAuth):
    "Authenticate with username and password. Applies superuser role as only permission"

    async def authenticate(self, request: Request, **kwargs: dict[str, str]):
        email: str = kwargs.get("username")  # type:ignore
        password: str = kwargs.get("password")  # type:ignore
        assert email
        assert password
        async with get_db() as db:
            user = await (
                await db.execute(f"SELECT * FROM users where email = '{email}'")
            ).fetchone()
        if not user:
            raise ValueError("No user found in database")
        if not auth.hash_password(password) == user["password"]:
            return None
        user_dict = dict(user)
        del user_dict["password"]
        auth_data = auth.AuthSessionData(
            is_authenticated=True,
            auth_handler="SuperuserPasswordAuth",
            user_id=user["id"],
            data=dict(user),
            permissions=["superuser"],
        )
        return auth_data

    async def get_user(self, user_id: int) -> auth.AuthSessionData:
        async with get_db() as db:
            user = await (
                await db.execute(
                    f"SELECT id, name, email, is_active FROM users where id = {user_id}"
                )
            ).fetchone()
        if not user:
            raise ValueError("No user found in database")
        return auth.AuthSessionData(
            is_authenticated=True,
            auth_handler="SuperuserPasswordAuth",
            user_id=user["id"],
            data=dict(user),
            permissions=["superuser"],
        )


auth.include_auth_handler(PasswordAuth, primary=True)
auth.include_auth_handler(SuperuserPasswordAuth)


class TokenAuth(auth.BaseAuth):
    async def authenticate(self, request: Request, **kwargs: dict[str, str]):
        token: str = kwargs.get("token")  # type:ignore
        # Doesn't do anything with the token. Just immitates another auth method
        result = await self.get_user(1)
        return result

    async def get_user(self, user_id: int) -> auth.AuthSessionData:
        password_auth = PasswordAuth()
        result = await password_auth.get_user(user_id)
        result["auth_handler"] = "TokenAuth"
        return result


auth.include_auth_handler(TokenAuth)


@protected_router.route("/login", methods=["GET", "POST"])
async def protected_login_route(request: Request, as_superuser: bool = False):
    if request.method == "POST":
        await auth.login(
            request,
            SuperuserPasswordAuth if as_superuser else PasswordAuth,
            username="test@email.com",
            password="password",
        )
        return f"logged in with {'SuperuserPasswordAuth' if as_superuser else 'PasswordAuth'}"
    return "login page"


@protected_router.route("/protected")
def protected_route():
    return "accessed"


@protected_router.route("/login-token", methods=["GET", "POST"])
async def login_token(request: Request):
    # Login using token authentication method
    if request.method == "POST":
        await auth.login(request, TokenAuth, token="random_token")
        return "token login success"
    return "token login page"


app.include_router(protected_router)


def test_route_protection():
    client.cookies.clear()
    result = client.get("/protected")
    assert result.status_code == 200  # Redirects to login page
    assert result.text != "accessed"
    assert result.text == "login page"
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/protected")
    assert result.status_code == 200
    assert result.text == "accessed"


def test_secondary_auth_method():
    print("testing secondary auth method")
    client.cookies.clear()
    result = client.get("/protected")
    assert result.status_code == 200
    assert result.text == "login page"
    result = client.post("/login-token")
    assert result.status_code == 200
    result = client.get("/protected")
    assert result.status_code == 200
    assert result.text == "accessed"


scope_protected_router = AppRouter()
scope_protected_router.add_middleware(auth.AuthMiddleware, allow_permissions=["admin"])


@scope_protected_router.route("/scope_protected_admin")
async def scope_protected_admin():
    return "scope protected admin"


invalid_scope_protected_router = AppRouter()
invalid_scope_protected_router.add_middleware(
    auth.AuthMiddleware, allow_permissions=["nope"]
)


@invalid_scope_protected_router.route("/invalid_scope_protected")
async def invalid_scope_protected():
    return "Invalid route"


app.include_router(scope_protected_router)
app.include_router(invalid_scope_protected_router)


def test_valid_scope_protected_router():
    client.cookies.clear()
    result = client.get("/scope_protected_admin")
    assert result.status_code == 200
    assert result.text == "login page"
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/scope_protected_admin")
    assert result.status_code == 200
    assert result.text == "scope protected admin"


def test_invalid_scope_protected_router():
    client.cookies.clear()
    result = client.get("/invalid_scope_protected")
    assert result.status_code == 200
    assert result.text == "login page"
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/invalid_scope_protected")
    assert result.status_code == 200
    assert result.text == "login page"


@app.route("/decorator_protected")
@auth.requires()
def decorator_protected_route(request: Request):
    return "decorator protected"


def test_decorator_protected():
    client.cookies.clear()  # Clear cookies
    result = client.get("/decorator_protected")
    assert result.status_code == 200  # Redirect to login page
    assert result.text != "decorator protected"
    assert result.text == "login page"
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/decorator_protected")
    assert result.status_code == 200
    assert result.text == "decorator protected"


@app.route("/decorator_protected_with_scope")
@auth.requires("admin")
def decorator_protected_with_scopes(request: Request):
    return "decorator protected with scope"


@app.route("/decorator_protected_missing_scope")
@auth.requires(["nope"])
def decorator_protected_missing_scope(request: Request, id: int):
    return "decorator protected missing scope"


def test_decorator_protected_with_scope():
    client.cookies.clear()  # Clear cookies
    result = client.get("/decorator_protected_with_scope")
    assert result.status_code == 200  # Redirect to login page
    assert result.text != "decorator protected with scope"
    assert result.text == "login page"
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/decorator_protected_with_scope")
    assert result.text == "decorator protected with scope"


def test_decorator_protected_missing_scope():
    client.cookies.clear()  # Clear cookies
    result = client.get("/decorator_protected_missing_scope")
    assert result.status_code == 200  # Redirect to login page
    assert result.text != "decorator protected missing scope"
    assert result.text == "login page"
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/decorator_protected_missing_scope")
    assert result.text != "decorator protected missing scope"
    assert result.text == "login page"


@app.route("/logout", methods=["POST"])
def logout(request: Request):
    auth.logout(request)


def test_logout():
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/protected")
    assert result.status_code == 200
    result = client.post("/logout")
    assert result.status_code == 200
    result = client.get("/protected")
    assert result.status_code == 200
    assert result.text == "login page"


@app.route("/superuser")
@auth.requires("nope")
def superuser_scope(request: Request):
    return "superuser"


def test_superuser_scope():
    result = client.post("/login", params={"as_superuser": True})
    assert result.status_code == 200
    assert result.text == "logged in with SuperuserPasswordAuth"
    result = client.get("/superuser")
    assert result.status_code == 200
    assert result.text == "superuser"


scope_admin_protected_router = AppRouter()
scope_admin_protected_router.add_middleware(
    auth.AuthMiddleware, allow_permissions=["admin"]
)


@scope_admin_protected_router.route("/scope_admin_protected_route")
def scope_admin_protected_route():
    return "scope_admin_protected_route"


scope_invalid_protected_router = AppRouter()
scope_invalid_protected_router.add_middleware(
    auth.AuthMiddleware, allow_permissions=["invalid"]
)


@scope_invalid_protected_router.route("/scope_invalid_protected_route")
def scope_invalid_protected_route():
    return "scope_invalid_protected_route"


app.include_router(scope_admin_protected_router)
app.include_router(scope_invalid_protected_router)


def test_auth_required_middleware_scopes():
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/scope_admin_protected_route")
    assert result.status_code == 200
    result = client.get("/scope_invalid_protected_route")
    assert result.status_code == 200
    assert result.text == "login page"  # Redirected to login page



from typing import Any

from mojito import (
    AppRouter,
    Jinja2Templates,
    JSONResponse,
    Mojito,
    Request,
    auth,
)

from .db import get_db

app = Mojito()

templates = Jinja2Templates("tests/templates")

main_router = AppRouter()


@main_router.route("/")
async def index():
    return "index_response"


@main_router.route("/async_route")
async def async_route():
    return "async_route_response"


@main_router.route("/{id:int}")
async def id_route_with_query_params(id: int, query_param_1: str, request: Request):
    return JSONResponse({"id": id, "query_param_1": query_param_1})


# TEST PROTECTED ROUTES
class PasswordAuth(auth.BaseAuth):
    "Authenticate with username and password"

    async def authenticate(self, request: Request, **kwargs: dict[str, str]):
        email: str = kwargs.get("username")  # type:ignore
        password: str = kwargs.get("password")  # type:ignore
        assert email
        assert password
        async with get_db() as db:
            user = await (
                await db.execute(f"SELECT * FROM users where email = '{email}'")
            ).fetchone()
        if not user:
            raise ValueError("No user found in database")
        if not auth.hash_password(password) == user["password"]:
            return None
        user_dict = dict(user)
        del user_dict["password"]
        auth_data = auth.AuthSessionData(
            is_authenticated=True,
            auth_handler="PasswordAuth",
            user_id=user["id"],
            data=dict(user),
            permissions=["admin"],
        )
        return auth_data

    async def get_user(self, user_id: Any) -> auth.AuthSessionData:
        async with get_db() as db:
            user = await (
                await db.execute(
                    f"SELECT id, name, email, is_active FROM users where id = {user_id}"
                )
            ).fetchone()
        if not user:
            raise ValueError("No user found in database")
        return auth.AuthSessionData(
            is_authenticated=True,
            auth_handler="PasswordAuth",
            user_id=user["id"],
            data=dict(user),
            permissions=["admin"],
        )


auth.include_auth_handler(PasswordAuth)

protected_subrouter = AppRouter("/protected")
protected_subrouter.add_middleware(auth.AuthMiddleware)


@protected_subrouter.route("/")
def protected_route():
    return "<p>protected</p>"


@app.route("/app-route")
def app_route():
    return "app-route"


main_router.include_router(protected_subrouter)

app.include_router(main_router)

if __name__ == "__main__":
    import uvicorn

    for route in app.routes:
        print(f"route: {route}")
    uvicorn.run("tests.main:app", reload=True)



from mojito import AppRouter, JSONResponse, Mojito, flash_message, get_flashed_messages
from mojito.testclient import TestClient

app = Mojito()
router = AppRouter()

client = TestClient(app)


@router.route("/set-flash")
def set_flash():
    flash_message("flash_set")
    flash_message("flash message 2", "warn")


@router.route("/get-flash")
def get_flash():
    return JSONResponse(get_flashed_messages())


app.include_router(router)


def test_message_flash():
    response = client.get("/set-flash")
    assert response.status_code == 200
    response = client.get("/get-flash")
    assert response.status_code == 200
    messages = response.json()
    assert "flash_set" == messages[0].get("message")
    assert "flash message 2" == messages[1].get("message")
    assert "warn" == messages[1].get("category")



from __future__ import annotations

import datetime
import json
import typing
from base64 import b64decode, b64encode

import itsdangerous
from itsdangerous.exc import BadSignature
from starlette.datastructures import MutableHeaders
from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from .. import config

if typing.TYPE_CHECKING:
    from ..auth import AuthSessionData


class UserSessionMiddleware:
    """Adds `user` data to the Request object.

    Used for authentication and session management.
    """

    # Works nearly the same as Starlettes SessionMiddleware but without some of the configuration options
    def __init__(
        self,
        app: ASGIApp,
        same_site: typing.Literal["lax", "strict", "none"] = "strict",
        https_only: bool = False,
        domain: str | None = None,
    ) -> None:
        self.app = app
        self.signer = itsdangerous.TimestampSigner(config.Config.SECRET_KEY)
        self.cookie_name = config.Config.USER_SESSION_COOKIE
        self.max_age = config.Config.USER_SESSION_EXPIRES
        self.path = "/"
        self.security_flags = "httponly; samesite=" + same_site
        if https_only:  # Secure flag can be used with HTTPS only
            self.security_flags += "; secure"
        if domain is not None:
            self.security_flags += f"; domain={domain}"

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):  # pragma: no cover
            await self.app(scope, receive, send)
            return

        connection = HTTPConnection(scope)
        initial_user_was_empty = True

        if self.cookie_name in connection.cookies:
            # RECEIVE COOKIE
            data = connection.cookies[self.cookie_name].encode("utf-8")
            try:
                data_bytes, timestamp = self.signer.unsign(
                    data, max_age=self.max_age, return_timestamp=True
                )
                data_decoded = b64decode(data_bytes)
                json_data: AuthSessionData = json.loads(data_decoded)
                reauthenticate_after = timestamp + datetime.timedelta(
                    seconds=config.Config.USER_SESSION_REVALIDATE_AFTER
                )
                if datetime.datetime.now(datetime.timezone.utc) > reauthenticate_after:
                    json_data["is_authenticated"] = (
                        False  # Force reauthentication if session needs revalidated
                    )
                scope["user"] = json_data
                initial_user_was_empty = False
            except BadSignature:
                scope["user"] = {}
        else:
            scope["user"] = {}

        async def send_wrapper(message: Message) -> None:
            # SEND COOKIE
            if message["type"] == "http.response.start":
                if scope["user"]:
                    # We have user data to persist.
                    data = b64encode(json.dumps(scope["user"]).encode("utf-8"))
                    data = self.signer.sign(data)
                    headers = MutableHeaders(scope=message)
                    header_value = "{cookie_name}={data}; path={path}; {max_age}{security_flags}".format(  # noqa E501
                        cookie_name=self.cookie_name,
                        data=data.decode("utf-8"),
                        path=self.path,
                        max_age=f"Max-Age={self.max_age}; "
                        if scope["user"].get("persist_session", False)
                        else "",
                        security_flags=self.security_flags,
                    )
                    headers.append("Set-Cookie", header_value)
                elif not initial_user_was_empty:
                    # The user has been cleared.
                    headers = MutableHeaders(scope=message)
                    header_value = "{cookie_name}={data}; path={path}; {expires}{security_flags}".format(  # noqa E501
                        cookie_name=self.cookie_name,
                        data="null",
                        path=self.path,
                        expires="expires=Thu, 01 Jan 1970 00:00:00 GMT; ",
                        security_flags=self.security_flags,
                    )
                    headers.append("Set-Cookie", header_value)
            await send(message)

        await self.app(scope, receive, send_wrapper)



from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Any, TypeVar, get_origin

from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema
from starlette.datastructures import FormData
from starlette.datastructures import UploadFile as StarletteUploadFile

from .requests import Request

try:
    from pydantic import BaseModel
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "form module requires pydantic being installed. \npip install pydantic"
    )


PydanticModel = TypeVar("PydanticModel", bound=BaseModel)


def _process_form(form: FormData, model: type[PydanticModel]) -> dict[str, Any]:
    # Preprocesses the form data before pydantic does any validation.
    # 1. Empty form inputs are sent as "" empty strings. Check and delete them from the
    #   form response before pydantic validates it
    # 2. Combine fields with the same name into a list of the fields values
    items = form.multi_items()
    fields = model.model_fields
    processed_items: dict[str, Any] = {}
    for item_name, item_value in items:
        # Build dict to return
        pydantic_field = fields.get(item_name)
        if isinstance(item_value, str) and not len(item_value) > 0:
            continue  # Skip this value if it is an empty string
        if processed_items.get(item_name):
            current_value = processed_items[item_name]
            # Item with that name already exists. Make or append to list.
            if isinstance(current_value, list):
                processed_items[item_name].append(item_value)
            else:
                processed_items[item_name] = [current_value, item_value]
            continue
        elif pydantic_field and get_origin(pydantic_field.annotation) is list:
            # Field doesn't exist yet and is defined as a list in the Pydantic model
            processed_items[item_name] = [item_value]
            continue
        processed_items[item_name] = item_value

    return processed_items


@asynccontextmanager
async def FormManager(
    request: Request,
    model: type[PydanticModel],
    max_files: int = 1000,
    max_fields: int = 1000,
) -> AsyncGenerator[PydanticModel, Any]:
    """Read form data from the request and validate it's content against a Pydantic model
    and return the valid Pydantic model. Extra data in the form is ignored and not passed into the
    Pydantic model. This does not work for processing files. You must use the request directly to get and read
    from files before using this function to read and validate the other form fields. See
    https://www.starlette.io/requests/#request-files for working with files.

    Args:
        request (Request): Mojito Request object
        model (PydanticModel): The Pydantic model to validate against
        max_files (int): The maximum number of files for Starlette to allow
        max_fields (int): The maximum number of fields for Starlette to allow

    Yields:
        PydanticModel: The validated Pydantic model


    Raises:
        ValidationError: Pydantic validation error
    """
    async with request.form(max_files=max_files, max_fields=max_fields) as form:
        valid_model = model.model_validate(_process_form(form, model))
        yield valid_model  # Yield result while in context of request.form()


async def Form(
    request: Request,
    model: type[PydanticModel],
    max_files: int = 1000,
    max_fields: int = 1000,
) -> PydanticModel:
    "Validates the form fields against the model"
    async with request.form(max_files=max_files, max_fields=max_fields) as form:
        valid_model = model.model_validate(_process_form(form, model))
        return valid_model


class UploadFile(StarletteUploadFile):
    """An uploaded file included as part of the request data.

    This is a subclass of starlette.datastructures.UploadFile that can be used in a Pydantic
    BaseModel class. Pydantic will pass the model through as-is without validation.
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        # Allow this file type to pass through pydantic without schema validation
        return core_schema.any_schema()



import contextlib

import aiosqlite

from mojito import auth


@contextlib.asynccontextmanager
async def get_db():
    # Creates an in-memory test database
    # Creates the admin user and yields a connection
    # async with aiosqlite.connect(":memory:") as conn:
    async with aiosqlite.connect("test.db") as conn:
        conn.row_factory = aiosqlite.Row
        cursor = await conn.cursor()
        await cursor.execute("""
CREATE TABLE IF NOT EXISTS users (
	id INTEGER PRIMARY KEY AUTOINCREMENT,
	name TEXT NOT NULL,
	email TEXT NOT NULL UNIQUE,
	password TEXT NOT NULL,
	is_active BOOL NOT NULL
);
""")
        existing_user = await (
            await cursor.execute('select * from users where email = "test@email.com"')
        ).fetchone()
        if not existing_user:
            await cursor.execute(
                f"""
    INSERT INTO users (name, email, password, is_active) VALUES
        (?, ?, ?, ?);                        
    """,
                ("Test User", "test@email.com", auth.hash_password("password"), True),
            )
        cursor = await conn.cursor()
        yield cursor



import os
from typing import Optional


class Config:
    SECRET_KEY: str = os.getenv("SECRET_KEY", "")
    """Secret key used to encrypt sensitive data.
    
    Defaults to empty string.
    """
    MESSAGE_FLASH_COOKIE: str = os.getenv("MESSAGE_FLASH_COOKIE", "mo_flash_messages")
    """Name of the cookie message flash data will be stored to.
    
    Defaults to `mo_flash_messages`
    """
    LOGIN_URL: str = os.getenv("LOGIN_URL", "/login")
    "Defaults to `/login`"
    USER_SESSION_COOKIE: str = os.getenv("USER_SESSION_COOKIE", "mo_user_session")
    """Name of the user session cookie.
    
    Defaults to `mo_user_session`
    """
    USER_SESSION_EXPIRES: int = int(os.getenv("USER_SESSION_EXPIRES", 60 * 60 * 24 * 7))
    """In seconds. How long the session cookie should be allowed to exist before being replaced.
    
    Defaults to 1 week.
    """
    USER_SESSION_REVALIDATE_AFTER: int = int(
        os.getenv("USER_SESSION_REVALIDATE_AFTER", 0)
    )
    """In seconds. How long the session cookie should be considered valid before running revalidation
    on the users authentication. Within this time, as long as the cookie signature is
    valid, the user will be considered authenticated and all the cookie data will be used.

    Defaults to 0, revalidate on every request.
    """
    SUPERUSER_PERMISSION_NAME: Optional[str] = os.getenv("SUPERUSER_PERMISSION_NAME")
    """The name of the superuser permission.

    When present in the Request.user.permissions, this role will bypass any permission (authorization) 
    requirements defined for a route function. This permission does not bypass authentication.
    """



import os

import pytest
from pydantic import BaseModel, Field, ValidationError

from mojito import JSONResponse, Mojito, Request, Response
from mojito.forms import Form, FormManager, UploadFile
from mojito.testclient import TestClient

app = Mojito()
client = TestClient(app)


class FormTest(BaseModel):
    field_1: str
    field_2: str
    field_3: int = Field(default=10)  # Optional field


@app.route("/", methods=["POST"])
async def process_form(request: Request):
    try:
        form = await Form(request, FormTest)
        return JSONResponse(form.model_dump())
    except ValidationError as e:
        return Response(e.__str__(), status_code=500)


@pytest.mark.parametrize(  # type: ignore
    ("form_data", "status"),
    [
        (
            {
                "field_1": "field one",
                "field_2": "field two",
                "field_3": 15,
            },
            200,
        ),
        (
            {
                "field_1": "field one",
                "field_2": "field two",
                "field_3": "nope",  # Invalid type
            },
            500,
        ),
        ({"field_1": "field one", "field_2": "field two"}, 200),
        (
            {
                "field_1": "field one",
                "field_2": "field two",
                "extra_field": "extra field",  # Ignored field
            },
            200,
        ),
        (
            {
                "field_1": "field one",  # Missing field_2
            },
            500,
        ),
        (
            {
                "field_1": "field one",
                "field_2": "",  # Missing field_2. Sent as empty string from form input
            },
            500,
        ),
    ],
)
def test_form_processing(form_data: dict[str, str], status: int):
    result = client.post("/", data=form_data)
    assert result.status_code == status
    if result.status_code == 200:
        assert result.json().get("field_1") == form_data.get("field_1")


class FormWithFileTest(BaseModel):
    field_1: str = ""
    file_1: UploadFile


@app.route("/upload_file", methods=["POST"])
async def process_upload_file_form(request: Request):
    async with FormManager(request, FormWithFileTest) as form:
        file_content = await form.file_1.read()
        return file_content.decode()


def test_form_file_processing():
    directory = os.path.dirname(__file__)
    file = open(directory + "/txt_file.txt", "rb")
    file_content = file.read()
    result = client.post("/upload_file", files={"file_1": file})
    assert result.status_code == 200
    assert result.text == file_content.decode()


class FormWithMultipleInputs(BaseModel):
    roles: list[str] = []
    other_input: str


@app.route("/checkboxes", methods=["POST"])
async def combine_checkboxes(request: Request):
    form = await Form(request, FormWithMultipleInputs)
    print(form.model_dump())
    return form.model_dump()


@pytest.mark.parametrize(
    ("form_data", "status"),
    [
        ({"roles": ["role 1", "role 2", "role 3"], "other_input": "input value"}, 200),
        (
            {"roles": "role 1", "other_input": "input value"},
            200,  # single role coerced to list
        ),
        (
            {"other_input": "input value"},
            200,  # role set as empty list
        ),
    ],
)
def test_form_combine_inputs(form_data: dict[str, str], status: int):
    result = client.post(
        "/checkboxes",
        data=form_data,
    )
    assert result.status_code == status



"""Mojito framework. Fresh, fast, and simple web framework for building HTML-first websites"""

__version__ = "0.2.0"

from .app import Mojito as Mojito
from .globals import g as g
from .helpers import (
    flash_message as flash_message,
)
from .helpers import (
    get_flashed_messages as get_flashed_messages,
)
from .requests import Request as Request
from .responses import (
    FileResponse as FileResponse,
)
from .responses import (
    HTMLResponse as HTMLResponse,
)
from .responses import (
    JSONResponse as JSONResponse,
)
from .responses import (
    PlainTextResponse as PlainTextResponse,
)
from .responses import (
    RedirectResponse as RedirectResponse,
)
from .responses import (
    Response as Response,
)
from .responses import (
    StreamingResponse as StreamingResponse,
)
from .routing import AppRouter as AppRouter
from .routing import redirect_to as redirect_to
from .staticfiles import StaticFiles as StaticFiles
from .templating import Jinja2Templates as Jinja2Templates



"""Authentication and authorization utilities to reduce the boilerplate required to implement basic session
based authentication."""

import functools
import hashlib
import sys
import typing as t

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec

from starlette.requests import Request
from starlette.responses import RedirectResponse
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from .config import Config


class AuthSessionData(t.TypedDict):
    """The authentication data stored on the user session."""

    is_authenticated: bool
    auth_handler: t.Optional[str]
    'The name of the auth handler class used to authenticate the user. Ex: "PasswordAuth"'
    user_id: t.Any
    data: dict[str, t.Any]
    """The user data dict. May contain any additional data about the user, such as name and tenant_id that
    you want to be available anywhere with access to the request. Don't store any sensitive
    information like passwords as all of this will be encoded and stored on the user session but may
    be decoded by anyone who inspects the cookie."""
    permissions: list[str]
    "Permissions are used to authorize the user. Think of them as scopes in a JWT."


class BaseAuth(t.Protocol):
    """Base class that all authentication methods should implement."""

    async def authenticate(
        self, request: Request, **kwargs: dict[str, t.Any]
    ) -> t.Optional[AuthSessionData]:
        """Method to authenticate the user based on the users username and password. Will
        be used by the password_login() function to authenticate the user.

        Args:
            request (Request): Mojito/Starlette request object
            **kwargs (P.kwargs): The credentials to use in authorization as keyword only arguments

        Raises:
            NotImplementedError: Method not implemented

        Returns:
            AuthSessionData | None: The auth data stored on the session.
        """
        raise NotImplementedError()

    async def get_user(self, user_id: t.Any) -> AuthSessionData:
        """Fetch the user based on user_id. Used when revalidating the user without requiring
        reauthentication.

        Implementation should get the latest permissions and verify the user is still active.

        Args:
            user_id (Any): Unique ID to locate the user in storage.
        """
        raise NotImplementedError()


class _AuthConfig:
    default_handler: t.Optional[str] = None
    "Name of the default hanlder class"
    auth_handlers: dict[str, type[BaseAuth]] = {}

    @staticmethod
    def get_default_handler() -> str:
        if not _AuthConfig.default_handler:
            raise NotImplementedError(
                "must set a default auth handler using set_auth_handler()"
            )
        return _AuthConfig.default_handler


def include_auth_handler(auth_handler: type[BaseAuth], primary: bool = False) -> None:
    """Include an auth handler for use. If this is the first or only auth handler included, it
    will be set as primary regardless of the value of the primary argument.

    Args:
        auth_handler (type[BaseAuth]): The class auth handler class.
        primary (bool, optional): Set the class as the primary/default auth handler. Defaults to False.
    """
    handler_name = auth_handler.__name__
    _AuthConfig.auth_handlers[handler_name] = auth_handler
    if primary or len(_AuthConfig.auth_handlers) == 1:
        _AuthConfig.default_handler = handler_name


async def _check_session_auth(request: Request, allowed_permissions: list[str]) -> bool:
    if not request.user:
        return False
    auth_session_data = AuthSessionData(
        is_authenticated=request.user.get("is_authenticated", False),
        auth_handler=request.user.get("auth_handler", _AuthConfig.default_handler),
        user_id=request.user.get("user_id"),
        data=request.user.get("user", {}),
        permissions=request.user.get("permissions", []),
    )
    if not auth_session_data["is_authenticated"]:
        # Revalidate on session exists but is not authenticated
        if not _AuthConfig.default_handler:
            raise NotImplementedError(
                "an auth handler must be set using set_auth_handler"
            )
        handler = _AuthConfig.auth_handlers[auth_session_data["auth_handler"]]  # type:ignore
        data = await handler().get_user(auth_session_data["user_id"])
        request.user.update(data)
    if (
        Config.SUPERUSER_PERMISSION_NAME
        and Config.SUPERUSER_PERMISSION_NAME in auth_session_data["permissions"]
    ):
        return True
    for allowed_permission in allowed_permissions:
        if not allowed_permission in auth_session_data["permissions"]:
            return False
    return True


class AuthMiddleware:
    """Uses sessions to authorize and authenticate users with requests.
    Redirect to login_url if session is not authenticated or if user does not have the required auth scopes.
    Can be applied at the app level or on individual routers.

    Will ignore the Config.LOGIN_URL path to prevent infinite redirects.

    Args:
        ignore_routes (Optional[list[str]]): defaults to None. paths of routes to ignore validation on like '/login'. Path should be relative
            and match the Request.url.path value when the route is called.
        allow_permissions (Optional[list[str]]): defaults to None. List of scopes the user must have in order to be authorized
            to access the requested resource.
    """

    def __init__(
        self,
        app: ASGIApp,
        ignore_routes: list[str] = [],
        allow_permissions: list[str] = [],
    ) -> None:
        self.app = app
        self.ignore_routes = ignore_routes
        self.allow_permissions = allow_permissions

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            return await self.app(scope, receive, send)
        request = Request(scope, receive)

        async def send_wrapper(message: Message) -> None:
            # ... Do something
            if (
                request.url.path in self.ignore_routes
                or request.url.path == Config.LOGIN_URL
            ):
                # Skip for routes registered as login_not_required
                return await send(message)
            allowed = await _check_session_auth(request, self.allow_permissions)
            if not allowed:
                response = RedirectResponse(Config.LOGIN_URL, 302)
                return await response(scope, receive, send)
            await send(message)

        await self.app(scope, receive, send_wrapper)


_P = ParamSpec("_P")


def requires(
    scopes: t.Union[str, t.Sequence[str]] = [],
    redirect_url: t.Optional[str] = None,
) -> t.Callable[[t.Callable[_P, t.Any]], t.Callable[_P, t.Any]]:
    """Decorator to require that the user is authenticated and optionally check that the user has
    the required auth scopes before accessing the resource. Redirect to the configured
    login_url if one is set, or to redirect_url if one is given.

    Args:
        scopes (str | Sequence[str]): Auth scopes to verify the user has. Defaults to [].
        redirect_url (Optional[str]): Redirect to this url rather than the configured
            login_url.
    """
    scopes_list = [scopes] if isinstance(scopes, str) else list(scopes)

    def decorator(
        func: t.Callable[_P, t.Any],
    ) -> t.Callable[..., t.Awaitable[t.Any]]:
        # Handle async request/response functions.
        @functools.wraps(func)
        async def wrapper(
            request: t.Union[Request, t.Any], *args: _P.args, **kwargs: _P.kwargs
        ) -> t.Any:
            if not isinstance(request, Request):
                raise Exception(
                    "The Request must be the first argument to the function when using the `@requires` decorator"
                )
            if not await _check_session_auth(request, scopes_list):
                REDIRECT_URL = redirect_url if redirect_url else Config.LOGIN_URL
                return RedirectResponse(REDIRECT_URL, 302)
            if isinstance(func, t.Awaitable):  # type: ignore [unused-ignore]
                return await func(request, *args, **kwargs)
            else:
                return func(request, *args, **kwargs)  # type: ignore

        return wrapper

    return decorator


def hash_password(password: str) -> str:
    """Helper to hash a password before storing it or to compare a plain text password to the one stored.

    Args:
        password (str): The plain text password to hash.

    Returns:
        str: The hashed password.
    """
    return hashlib.sha256(password.encode()).hexdigest()


async def login(
    request: Request,
    auth_handler: t.Optional[type[BaseAuth]] = None,
    persist_session: bool = True,
    **kwargs: t.Any,
) -> t.Optional[AuthSessionData]:
    """Login user and create an authenticated session.

    Args:
        request (Request): the request
        auth_handler (type[BaseAuth], optional): Auth class to login with.
            Defaults to the default auth handler configured using set_auth_handler().
        persist_session (bool): maintain the session cookie until the maximum age defined
            in the USER_SESSION_EXPIRES environment variable. If false, the cookie is set to
            expire at the end of the browser session. Default True.

    Returns:
        AuthSessionData: The data set on the session. None if invalid login.
    """
    if not auth_handler:
        auth_handler = _AuthConfig.auth_handlers[_AuthConfig.get_default_handler()]
    handler = auth_handler()
    result = await handler.authenticate(request, **kwargs)
    if not result:
        return None
    request.user.update(result)
    if not persist_session:
        request.user["persist_session"] = True
    return result


def logout(request: Request) -> None:
    """Expire the current user session."""
    request.user.clear()
